Attribute VB_Name = "modDelaunay2D"
'=================================================================
' Descrizione.....: Routines per la triangolazione Delaunay in 2D.
' Nome dei Files..: modDelaunay2D.bas, modUtilita
' Data............: 28/03/2003
' Versione........: 1.0 a 32 bits.
' Sistema.........: VB6 (SP5) sotto Windows NT (SP6).
' Scritto da......: F. Languasco 
' E-Mail..........: MC7061@mclink.it
' DownLoads a.....: http://www.flanguasco.org
'                   http://members.xoom.virgilio.it/flanguasco/
'=================================================================
'
'   Routines di ingresso:
'    DTRIS2 NPT, VCL(), IND(), NTRI, TIL(), TNBR(), IERR
'    DTRIW2 NPT, VCL(), IND(), NTRI, TIL(), TNBR(), IERR
'
'    Parametri: vedi descrizione all' interno delle routines.
'
'    Nota:  Alla chiamata e' necessario che vengano date le dimensioni
'           TIL(1 To 3, 1 To NPT * 2) e TNBR(1 To 3, 1 To NPT * 2);
'           All' uscita sono validi solo i valori TIL(1 To 3, 1 To NTRI)
'           e TNBR(1 To 3, 1 To NTRI).
'
'    Nota:  Tutti i vettori e le matrici di queste routines
'           iniziano dall' indice 1.
'
'    Nota:  Tutte le routines di questo modulo sono state tradotte
'           dal FORTRAN della libreria Geompack3.
'
Option Explicit
'
Private Const Epsilon# = 2# ^ -52 ' Epsilon del FORTRAN per numeri in doppia precisione.

Private Sub zCodiciErrore()
'
'
'   Codici di errore originali della libreria GEOMPACK3:
'
'   Error codes: Error codes IERR from 1 to 99 indicate not enough space in some array.
'   Error codes IERR >= 100 mean that the input specification is incorrect,
'    there is a program bug, or the tolerance TOL is too small or large (try
'   calling INITCB with a different TOLIN value).
'
'   IERR    DESCRIPTION OF ERROR
'     1  -  not enough space in EDGE array for routine EDGHT
'     2  -  not enough space in HOLV array for routine DSMCPR or DSPGDC
'     3  -  not enough space in 2-D VCL array
'     4  -  not enough space in HVL array
'     5  -  not enough space in PVL, IANG arrays
'     6  -  not enough space in IWK array
'     7  -  not enough space in WK array
'     8  -  not enough space in STACK array for routine SWAPEC, DTRIS2, or DTRIW2
'     9  -  not enough space in TIL array
'    10  -  not enough space in CWALK array for routine INTTRI
'    11  -  not enough space in FC array for 3-D
'    12  -  not enough space in BF array for 3-D
'    13  -  not enough space in SVCL, SFVL arrays for routine SHRNK3
'    14  -  not enough space in 3-D VCL array
'    15  -  not enough space in FVL, EANG arrays for polyhedral decomposition
'    16  -  not enough space in FACEP, NRML arrays
'    17  -  not enough space in PFL array
'    18  -  not enough space in HFL array
'    19  -  not enough space in BTL array
'    20  -  not enough space in VM array
'    21  -  not enough space in HT array
'    22  -  not enough space in FC array for K-D
'    23  -  not enough space in BF array for K-D
'   100  -  higher primes need to be added to routine PRIME
'   200  -  abnormal return in routine DIAM2
'   201  -  abnormal return in routine WIDTH2
'   202  -  parallel lines in routine SHRNK2
'   205  -  horizontal ray to right does not intersect polygon in routine ROTIPG
'   206  -  out-of-range index when popping points from stack in routine VPRGHT
'   207  -  out-of-range index when scanning vertices in routine VPSCNA
'   208  -  out-of-range index when scanning vertices in routine VPSCNB
'   209  -  out-of-range index when scanning vertices in routine VPSCNC
'   210  -  out-of-range index when scanning vertices in routine VPSCND
'   212  -  singular matrix in routine VORNBR
'   215  -  unmatched separator interface edge determined by routine DSPGDC
'   216  -  all edges of an outer boundary curve are specified as hole
'           edges in input to routine DSPGDC
'   218  -  cannot find subregion above top vertex of hole in routine JNHOLE
'   219  -  angle at hole vertex in modified region is too far from PI due
'           to use of relative tolerance in routine JNHOLE
'   222  -  separator not found in routine MFDEC2 (not likely to occur if
'           ANGSPC <= 30 degrees and ANGTOL <= 20 degrees)
'   224  -  2 vertices with identical coordinates (in floating point arithmetic)
'           detected in routine DTRIS2 or DTRIW2
'   225  -  all vertices input to DTRIS2 or DTRIW2 are collinear (in fl pt arith)
'   226  -  cycle detected in walk by routine WALKT2
'   230  -  invalid (CW-oriented) triangle created in routine TMERGE
'   231  -  unsuccessful search in routine FNDTRI
'   300  -  unsuccessful search by routine HTSRC
'   301  -  degenerate tetrahedron detected by routine BARYTH, CCSPH, or OPSIDE
'   302  -  2 vertices with identical coordinates (in floating point arithmetic)
'           detected in routine FRSTET, DTRIS3, DTRIW3, or ITRIS3
'   303  -  all vertices input to DTRIS3 or DTRIW3 are collinear (in fl pt arith)
'   304  -  all vertices input to DTRIS3 or DTRIW3 are coplanar (in fl pt arith)
'   305  -  invalid configuration (due to tolerance) detected by routine SWAPES
'   306  -  no visible boundary face found in routine DTRIS3 or ITRIS3
'   307  -  cycle detected in walk by routine WALKT3
'   308  -  nontransformable or nonexistent face at top of stack in routine SWAPTF
'   309  -  ALPHA(4) = 0 occurs after calling BARYTH in SWAPES, SWAPMU, etc.;
'           TOL must be decreased
'   310  -  unmatched edge determined by routine DSCPH
'   313  -  incorrect value for variable T in routine XPGHPL
'   314  -  unmatched edge in shrunken polyhedron in routine SHRNK3
'   315  -  2 "identical" vertices with unequal coordinates in routine SHRNK3
'   318  -  vertex ITOP not found in FVL array in routine INTMVG
'   321  -  face oriented same way twice in routine DSPHDC or DSPHFH
'   322  -  unmatched edge determined by routine DSPHDC, DSPHFH, or DSPHIH
'   325  -  unsuccessful search for next edge of cut polygon in routine CUTFAC
'   326  -  unsuccessful search for intersecting edge of face FL in CUTFAC
'   327  -  at least one reflex edge has not been resolved by routine RESEDG
'   328  -  value of constant MAXEV must be increased in routine CUTFAC
'   330  -  value of MAXSV must be increased before calling SHRNK3 in TRIPR3
'   331  -  invalid vertex detected from walk or linear search in routine BCDTRI
'   334  -  unsuccessful search for next edge of cut polygon in routine SEPFAC
'   335  -  unsuccessful search for intersecting edge of face FL in SEPFAC
'   336  -  separator face not found in routine MFDEC3
'   340  -  holes are not in nondecreasing face index order in routine DSPHFH
'   341  -  hole on non-boundary face of polyhedral region detected by DSPHFH
'   342  -  face containing non-coplanar vertices detected by DSPHFH
'   344  -  cannot find subregion above top (or below bottom) vertex of hole in
'           routine SPDECF
'   346  -  hole polyhedron not connected to outer boundary by routine DSPHIH
'   347  -  starting edge not found by routine RESHOL (because PT not in polyh)
'   348  -  starting edge not found by routine RESHOL (because ray meets vertex)
'   349  -  unsuccessful search for intersecting edge of face FMIN in RESHOL
'   400  -  unsuccessful search by routine HTSRCK or HTSDLK
'   401  -  degenerate simplex detected by routine BARYCK
'   402  -  2 vertices with identical coordinates (in floating point arithmetic)
'           detected in routine FRSMPX, DTRISK, DTRIWK, or DTRIMK
'   403  -  all vertices input to DTRISK, DTRIWK, or DTRIMK are in same hyperplane
'           (in floating point arithmetic)
'   405  -  invalid configuration (due to tolerance) detected by routine SWAPHS
'   406  -  no visible boundary face found in routine DTRISK
'   407  -  cycle detected in walk by routine WALKTK
'
'
'
End Sub
Private Sub WALKT2(ByVal X#, ByVal Y#, ByVal NTRI&, Vcl#(), _
    TIL&(), TNBR&(), ByRef ITRI&, ByRef IEDG&, ByRef IERR&)
'
'******************************************************************************
'
'   WALKT2 walks through a 2D triangulation searching for a point.
'
'
'   Purpose:
'
'    Walk through neighboring triangles of 2D (Delaunay)
'    triangulation until a triangle is found containing point (X,Y)
'    or (X,Y) is found to be outside the convex hull. Search is
'    guaranteed to terminate for a Delaunay triangulation, else a
'    cycle may occur.
'
'   Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'   Parameters:
'
'    Input, X, Y - 2D point.
'
'    Input, NTRI - number of triangles in triangulation; used to detect cycle.
'
'    Input, VCL(1:2,1:*) - coordinates of 2D vertices.
'
'    Input, TIL(1:3,1:*) - triangle incidence list.
'
'    Input, TNBR(1:3,1:*) - triangle neighbor list.
'
'    Input/output, ITRI.  On input, the index of triangle to begin search at.
'    On output, the index of triangle that search ends at.
'
'    Output, IEDG - 0 if (X,Y) is in the interior of triangle ITRI; I = 1,
'    2, or 3 if (X,Y) is on interior of edge I of ITRI;
'    I = 4, 5, or 6 if (X,Y) is (nearly) vertex I-3 of ITRI;
'    I = -1, -2, or -3 if (X,Y) is outside convex hull due
'    to walking past edge -I of triangle ITRI.
'
'    Output, integer IERR, error flag, which is zero unless an error occurred.
'
    Dim I&, a&, b&, c&, cnt&
    Dim Alpha#, Beta#, Gamma#, det#, Dx#, dXa#, dXb#, Dy#, dYa#, dYb#, tol#
'
'    Use barycentric coordinates to determine where (X,Y) is located.
'
    IERR = 0
    tol = 100# * Epsilon '(Tol)
    cnt = 0
    IEDG = 0
'
10  'CONTINUE
'
    cnt = cnt + 1
'
    If (cnt > NTRI) Then
            IERR = 226
            Exit Sub
    End If
'
    a = TIL(1, ITRI)
    b = TIL(2, ITRI)
    c = TIL(3, ITRI)
    dXa = Vcl(1, a) - Vcl(1, c)
    dYa = Vcl(2, a) - Vcl(2, c)
    dXb = Vcl(1, b) - Vcl(1, c)
    dYb = Vcl(2, b) - Vcl(2, c)
    Dx = X - Vcl(1, c)
    Dy = Y - Vcl(2, c)
    det = dXa * dYb - dYa * dXb
    Alpha = (Dx * dYb - Dy * dXb) / det
    Beta = (dXa * Dy - dYa * Dx) / det
    Gamma = 1# - Alpha - Beta
'
    If (Alpha > tol And Beta > tol And Gamma > tol) Then
        Exit Sub
'
    ElseIf (Alpha < -tol) Then
        I = TNBR(2, ITRI)
        If (I <= 0) Then
            IEDG = -2
            Exit Sub
        End If
'
    ElseIf (Beta < -tol) Then
        I = TNBR(3, ITRI)
        If (I <= 0) Then
            IEDG = -3
            Exit Sub
        End If
'
    ElseIf (Gamma < -tol) Then
        I = TNBR(1, ITRI)
        If (I <= 0) Then
            IEDG = -1
            Exit Sub
        End If
'
    ElseIf (Alpha <= tol) Then
        If (Beta <= tol) Then
            IEDG = 6
        ElseIf (Gamma <= tol) Then
            IEDG = 5
        Else
            IEDG = 2
        End If
        Exit Sub
'
    ElseIf (Beta <= tol) Then
        If (Gamma <= tol) Then
            IEDG = 4
        Else
            IEDG = 3
        End If
        Exit Sub
'
    Else
        IEDG = 1
        Exit Sub
    End If

    ITRI = I
    GoTo 10
'
'
'
End Sub
Public Sub DTRIW2(ByVal NPT&, Vcl#(), Ind&(), _
    ByRef NTRI&, ByRef TIL&(), ByRef TNBR&(), ByRef IERR&)
'
'******************************************************************************
'
'   DTRIW2 constructs a Delaunay triangulation of vertices in 2D.
'
'
'   Purpose:
'
'    Construct Delaunay triangulation of 2D vertices using
'    incremental approach and diagonal edge swaps. Vertices are
'    inserted one at a time in order given by IND array. The initial
'    triangles created due to a new vertex are obtained by a walk
'    through the triangulation until location of vertex is known.
'
'   Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'   Parameters:
'
'    Input, NPT - number of 2D points (vertices).
'
'    Input, MAXST - maximum size available for STACK array; should be about
'    NPT to be safe, but MAX(10,2*LOG2(NPT)) usually enough.
'
'    Input, VCL(1:2,1:*) - coordinates of 2D vertices.
'
'    Input, IND(1:NPT) - indices in VCL of vertices to be triangulated;
'    vertices are inserted in order given by this array
'
'    Output, NTRI - number of triangles in triangulation; equal to
'    2*NPT - NB - 2 where NB = number of boundary vertices.
'
'    Output, TIL(1:3,1:NTRI) - triangle incidence list; elements are indices
'    of VCL; vertices of triangles are in counter clockwise order.
'
'    Output, TNBR(1:3,1:NTRI) - triangle neighbor list; positive elements
'    are indices of TIL; negative elements are used for links
'    of counter clockwise linked list of boundary edges; LINK = -(3*I + J-1)
'    where I, J = triangle, edge index; TNBR(J,I) refers to
'    the neighbor along edge from vertex J to J+1 (mod 3).
'
'    Workspace, STACK(1:MAXST) - used for stack of triangles for which
'    circumcircle test must be made
'
'    Output, integer IERR, error flag, which is zero unless an error occurred.
'
    Dim cmax#, tol#
    Dim I&, I3&, J&, e&, EM1&, EP1&, M&, M1&, M2&, M3&, MAXST&, N&
    Dim L&, LEDG&, LR&, LTRI&, REDG&, RTRI&, T&, Top&
'
    MAXST = NPT
    ReDim STACK&(1 To MAXST)
'
    Const msglvl& = 0
'
'    Determine initial triangle.
'
    IERR = 0
    tol = 100# * Epsilon '(Tol)
'
    M1 = Ind(1)
    M2 = Ind(2)
'
    For J = 1 To 2
        cmax = DMAX1(Abs(Vcl(J, M1)), Abs(Vcl(J, M2)))
        If (Abs(Vcl(J, M1) - Vcl(J, M2)) > tol * cmax And cmax > tol) Then
            GoTo 20
        End If
    Next J
'
    IERR = 224
    Exit Sub
20  'CONTINUE
'
    I3 = 3
30  'CONTINUE
'
    If (I3 > NPT) Then
        IERR = 225
        Exit Sub
    End If
'
    M = Ind(I3)
    LR = LRLINE(Vcl(1, M), Vcl(2, M), Vcl(1, M1), Vcl(2, M1), Vcl(1, M2), _
         Vcl(2, M2), 0#)
'
    If (LR = 0) Then
        I3 = I3 + 1
        GoTo 30
    End If
'
    If (I3 <> 3) Then
        Ind(I3) = Ind(3)
        Ind(3) = M
    End If
'
    NTRI = 1
'
    If (LR = -1) Then
        TIL(1, 1) = M1
        TIL(2, 1) = M2
    Else
        TIL(1, 1) = M2
        TIL(2, 1) = M1
    End If
'
    TIL(3, 1) = M
    TNBR(1, 1) = -4
    TNBR(2, 1) = -5
    TNBR(3, 1) = -3
'
'    If (msglvl = 4) Then
'        write ( *,600) 1,vcl(1,m1),vcl(2,m1),vcl(1,m2),vcl(2,m2)
'        write ( *,600) 1,vcl(1,m2),vcl(2,m2),vcl(1,m),vcl(2,m)
'        write ( *,600) 1,vcl(1,m),vcl(2,m),vcl(1,m1),vcl(2,m1)
'    End If
'
'    Insert vertices one at a time from anywhere, walk through
'    triangulation to determine location of new vertex, and apply
'    diagonal edge swaps until Delaunay triangulation of vertices
'    (so far) is obtained.
'
    Top = 0
'
    For I = 4 To NPT
'        If (msglvl = 4) Then
'            write ( *,600) i
'        End If
'
        M = Ind(I)
        RTRI = NTRI
        Call WALKT2(Vcl(1, M), Vcl(2, M), NTRI, Vcl, TIL, TNBR, RTRI, REDG, IERR)
'
        If (REDG = 0) Then
            M1 = TIL(1, RTRI)
            M2 = TIL(2, RTRI)
            M3 = TIL(3, RTRI)
            TIL(3, RTRI) = M
'
            If (TNBR(1, RTRI) > 0) Then
                Top = 1
                STACK(Top) = RTRI
            End If
'
            NTRI = NTRI + 1
            TIL(1, NTRI) = M2
            TIL(2, NTRI) = M3
            TIL(3, NTRI) = M
            N = TNBR(2, RTRI)
            TNBR(1, NTRI) = N
'
            If (N > 0) Then
                If (TNBR(1, N) = RTRI) Then
                    TNBR(1, N) = NTRI
                ElseIf (TNBR(2, N) = RTRI) Then
                    TNBR(2, N) = NTRI
                Else
                    TNBR(3, N) = NTRI
                End If
'
                Top = Top + 1
                STACK(Top) = NTRI
            End If
'
            NTRI = NTRI + 1
            TIL(1, NTRI) = M3
            TIL(2, NTRI) = M1
            TIL(3, NTRI) = M
            N = TNBR(3, RTRI)
            TNBR(1, NTRI) = N
'
            If (N > 0) Then
                If (TNBR(1, N) = RTRI) Then
                    TNBR(1, N) = NTRI
                ElseIf (TNBR(2, N) = RTRI) Then
                    TNBR(2, N) = NTRI
                Else
                    TNBR(3, N) = NTRI
                End If
'
                Top = Top + 1
                STACK(Top) = NTRI
            End If
'
            TNBR(2, RTRI) = NTRI - 1
            TNBR(3, RTRI) = NTRI
            TNBR(2, NTRI - 1) = NTRI
            TNBR(3, NTRI - 1) = RTRI
            TNBR(2, NTRI) = RTRI
            TNBR(3, NTRI) = NTRI - 1
'
            If (TNBR(1, NTRI - 1) <= 0) Then
                T = RTRI
                e = 1
'
40              'CONTINUE
'
                If (TNBR(e, T) > 0) Then
                    T = TNBR(e, T)
'
                    If (TIL(1, T) = M2) Then
                        e = 3
                    ElseIf (TIL(2, T) = M2) Then
                        e = 1
                    Else
                        e = 2
                    End If
'
                    GoTo 40
                End If
'
                TNBR(e, T) = -3 * NTRI + 3
            End If
'
            If (TNBR(1, NTRI) <= 0) Then
                T = NTRI - 1
                e = 1
'
50              'CONTINUE
'
                If (TNBR(e, T) > 0) Then
                    T = TNBR(e, T)
'
                    If (TIL(1, T) = M3) Then
                        e = 3
                    ElseIf (TIL(2, T) = M3) Then
                        e = 1
                    Else
                        e = 2
                    End If
'
                    GoTo 50
                End If
'
                TNBR(e, T) = -3 * NTRI
'
            End If
'
'            if (msglvl = 4) then
'                write ( *,600) 1,vcl(1,m),vcl(2,m),vcl(1,m1),vcl(2,m1)
'                write ( *,600) 1,vcl(1,m),vcl(2,m),vcl(1,m2),vcl(2,m2)
'                write ( *,600) 1,vcl(1,m),vcl(2,m),vcl(1,m3),vcl(2,m3)
'            End If
'
        ElseIf (REDG < 0) Then
            REDG = -REDG
            LTRI = 0
'
            Call VBEDG(Vcl(1, M), Vcl(2, M), Vcl, TIL, TNBR, LTRI, LEDG, RTRI, _
                REDG)
'
            N = NTRI + 1
            L = -TNBR(LEDG, LTRI)
'
60          'CONTINUE
'
            T = L \ 3
            e = (L Mod 3) + 1
            L = -TNBR(e, T)
            M2 = TIL(e, T)
'
            If (e <= 2) Then
                M1 = TIL(e + 1, T)
            Else
                M1 = TIL(1, T)
            End If
'
            NTRI = NTRI + 1
            TNBR(e, T) = NTRI
            TIL(1, NTRI) = M1
            TIL(2, NTRI) = M2
            TIL(3, NTRI) = M
            TNBR(1, NTRI) = T
            TNBR(2, NTRI) = NTRI - 1
            TNBR(3, NTRI) = NTRI + 1
            Top = Top + 1
'
            If (Top > MAXST) Then
                IERR = 8
                GoTo 100
            End If
'
            STACK(Top) = NTRI
'
'            If (msglvl = 4) Then
'                write ( *,600) 1,vcl(1,m),vcl(2,m),vcl(1,m2),vcl(2,m2)
'            End If
'
            If (T <> RTRI Or e <> REDG) Then GoTo 60
'
'            If (msglvl = 4) Then
'                write ( *,600) 1,vcl(1,m),vcl(2,m), vcl(1,m1),vcl(2,m1)
'            End If
'
            TNBR(LEDG, LTRI) = -3 * N - 1
            TNBR(2, N) = -3 * NTRI - 2
            TNBR(3, NTRI) = -L
'
        ElseIf (REDG <= 3) Then
            M1 = TIL(REDG, RTRI)
'
            If (REDG = 1) Then
                e = 2
                EP1 = 3
            ElseIf (REDG = 2) Then
                e = 3
                EP1 = 1
            Else
                e = 1
                EP1 = 2
            End If
'
            M2 = TIL(e, RTRI)
            TIL(e, RTRI) = M
            M3 = TIL(EP1, RTRI)
'
            If (TNBR(EP1, RTRI) > 0) Then
                Top = 1
                STACK(Top) = RTRI
            End If
'
            NTRI = NTRI + 1
            TIL(1, NTRI) = M
            TIL(2, NTRI) = M2
            TIL(3, NTRI) = M3
            N = TNBR(e, RTRI)
            TNBR(2, NTRI) = N
            TNBR(3, NTRI) = RTRI
            TNBR(e, RTRI) = NTRI
'
            If (N > 0) Then
                If (TNBR(1, N) = RTRI) Then
                    TNBR(1, N) = NTRI
                ElseIf (TNBR(2, N) = RTRI) Then
                    TNBR(2, N) = NTRI
                Else
                    TNBR(3, N) = NTRI
                End If
'
                Top = Top + 1
                STACK(Top) = NTRI
            End If
'
'            If (msglvl = 4) Then
'                write ( *,600) 1,vcl(1,m),vcl(2,m),vcl(1,m3),vcl(2,m3)
'            End If
'
            LTRI = TNBR(REDG, RTRI)
'
            If (LTRI <= 0) Then
                TNBR(1, NTRI) = LTRI
                TNBR(REDG, RTRI) = -3 * NTRI
                If (TNBR(2, NTRI) <= 0) Then TNBR(1, NTRI) = -3 * NTRI - 1
'
            Else
                TNBR(1, NTRI) = NTRI + 1
                TNBR(REDG, RTRI) = LTRI
'
                If (TIL(1, LTRI) = M2) Then
                    LEDG = 1
                    EM1 = 2
                    e = 3
                ElseIf (TIL(2, LTRI) = M2) Then
                    LEDG = 2
                    EM1 = 3
                    e = 1
                Else
                    LEDG = 3
                    EM1 = 1
                    e = 2
                End If
'
                TIL(LEDG, LTRI) = M
                M3 = TIL(e, LTRI)
'
                If (TNBR(EM1, LTRI) > 0) Then
                    Top = Top + 1
                    STACK(Top) = LTRI
                End If
'
                NTRI = NTRI + 1
                TIL(1, NTRI) = M2
                TIL(2, NTRI) = M
                TIL(3, NTRI) = M3
                TNBR(1, NTRI) = NTRI - 1
                TNBR(2, NTRI) = LTRI
                N = TNBR(e, LTRI)
                TNBR(3, NTRI) = N
                TNBR(e, LTRI) = NTRI
'
                If (N > 0) Then
                    If (TNBR(1, N) = LTRI) Then
                        TNBR(1, N) = NTRI
                    ElseIf (TNBR(2, N) = LTRI) Then
                        TNBR(2, N) = NTRI
                    Else
                        TNBR(3, N) = NTRI
                    End If
'
                    Top = Top + 1
                    STACK(Top) = NTRI
                End If
'
'                If (msglvl = 4) Then
'                    write ( *,600) 1,vcl(1,m),vcl(2,m), vcl(1,m3),vcl(2,m3)
'                End If
'
                If (TNBR(2, NTRI - 1) <= 0) Then
                    T = NTRI
                    e = 3
'
70                  'CONTINUE
'
                    If (TNBR(e, T) > 0) Then
                        T = TNBR(e, T)
'
                        If (TIL(1, T) = M2) Then
                            e = 3
                        ElseIf (TIL(2, T) = M2) Then
                            e = 1
                        Else
                            e = 2
                        End If
'
                        GoTo 70
                    End If
'
                    TNBR(e, T) = -3 * NTRI + 2
                End If
'
                If (TNBR(3, NTRI) <= 0) Then
                    T = LTRI
'
                    If (LEDG <= 2) Then
                        e = LEDG + 1
                    Else
                        e = 1
                    End If
'
80                  'CONTINUE
'
                    If (TNBR(e, T) > 0) Then
                        T = TNBR(e, T)
'
                        If (TIL(1, T) = M3) Then
                            e = 3
                        ElseIf (TIL(2, T) = M3) Then
                            e = 1
                        Else
                            e = 2
                        End If
'
                        GoTo 80
'
                    End If
'
                    TNBR(e, T) = -3 * NTRI - 2
                End If
            End If
'
        Else
            IERR = 224
            GoTo 100
        End If
'
        Call SWAPEC(M, Top, MAXST, 0, 0, Vcl, TIL, TNBR, STACK, IERR)
'
        If (IERR <> 0) Then
            Exit For
        End If
    Next I
'
100 'CONTINUE
'
    If (I3 <> 3) Then
        T = Ind(I3)
        Ind(I3) = Ind(3)
        Ind(3) = T
    End If
'
'    if (msglvl = 4) then
'        write ( *,600) npt+1
'    End If

'600 format (1x,i7,4f15.7)
'
'
'
End Sub
Private Function DIAEDG(ByVal X0#, ByVal Y0#, ByVal X1#, ByVal Y1#, _
        ByVal X2#, ByVal Y2#, ByVal X3#, ByVal Y3#) As Long
'
'******************************************************************************
'
'   DIAEDG determines which diagonal to use in a quadrilateral.
'
'
'   Purpose:
'
'    Determine whether 02 or 13 is the diagonal edge chosen
'    based on the circumcircle criterion, where (X0,Y0), (X1,Y1),
'    (X2,Y2), (X3,Y3) are the vertices of a simple quadrilateral
'    in counterclockwise order.
'
'   Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada    T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'   Parameters:
'
'    Input, X0, Y0, X1, Y1, X2, Y2, X3, Y3 - vertex coordinates.
'
'    Output, DIAEDG -
'     1, if diagonal edge 02 is chosen, i.e. 02 is inside.
'        quadrilateral + vertex 3 is outside circumcircle 012
'    -1, if diagonal edge 13 is chosen, i.e. 13 is inside.
'        quadrilateral + vertex 0 is outside circumcircle 123
'     0, if four vertices are cocircular.
'
    Dim ca#, cb#, dx10#, dx12#, dx30#, dx32#, dy10#, dy12#, dy30#, dy32#
    Dim s#, tol#, tola#, tolb#
'
    tol = 100# * Epsilon '(tol)
    dx10 = X1 - X0
    dy10 = Y1 - Y0
    dx12 = X1 - X2
    dy12 = Y1 - Y2
    dx30 = X3 - X0
    dy30 = Y3 - Y0
    dx32 = X3 - X2
    dy32 = Y3 - Y2
    tola = tol * DMAX1(Abs(dx10), Abs(dy10), Abs(dx30), Abs(dy30))
    tolb = tol * DMAX1(Abs(dx12), Abs(dy12), Abs(dx32), Abs(dy32))
    ca = dx10 * dx30 + dy10 * dy30
    cb = dx12 * dx32 + dy12 * dy32
'
    If (ca > tola And cb > tolb) Then
        DIAEDG = -1
    ElseIf (ca < -tola And cb < -tolb) Then
        DIAEDG = 1
    Else
        tola = DMAX1(tola, tolb)
        s = (dx10 * dy30 - dx30 * dy10) * cb + (dx32 * dy12 - dx12 * dy32) * ca

        If (s > tola) Then
            DIAEDG = -1
        ElseIf (s < -tola) Then
            DIAEDG = 1
        Else
            DIAEDG = 0
        End If

    End If
'
'
'
End Function
Private Sub SWAPEC(ByVal I&, ByRef Top&, ByVal MAXST&, ByRef BTRI&, ByRef BEDG&, _
      Vcl#(), ByRef TIL&(), ByRef TNBR&(), STACK&(), ByRef IERR&)
'
'******************************************************************************
'
'   SWAPEC swaps diagonal edges in a 2D triangulation
'
'
'   Purpose:
'
'    Swap diagonal edges in 2D triangulation based on empty
'    circumcircle criterion until all triangles are Delaunay, given
'    that I is index of new vertex added to triangulation.
'
'   Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'   Parameters:
'
'    Input, I - index in VCL of new vertex.
'
'    Input/output, TOP - index of top of stack, >= 0.
'
'    Input, MAXST - maximum size available for STACK array.
'
'    Input/output, BTRI, BEDG - if positive, these are triangle and edge
'    index of a boundary edge whose updated indices must be recorded.
'
'    Input, VCL(1:2,1:*) - coordinates of 2D vertices.
'
'    Input/output, TIL(1:3,1:*) - triangle incidence list.
'
'    Input/output, TNBR(1:3,1:*) - triangle neighbor list; negative values are
'    used for links of counter clockwise linked list of boundary edges;
'    LINK = -(3*I + J-1) where I, J = triangle, edge index.
'
'    Input, STACK(1:TOP) - index of initial triangles (involving vertex I)
'    put in stack; the edges opposite I should be in interior.
'
'    Workspace, STACK(TOP+1:MAXST) - used as stack.
'
'    Output, integer IERR, error flag, which is zero unless an error occurred.
'
    Dim L&, a&, b&, c&, e&, ee&, EM1&, EP1&
    Dim F&, fm1&, fp1&, R&, s&, swap&, T&, tt&, u&
    Dim X#, Y#
'
    Const msglvl& = 0
'
'    Determine whether triangles in stack are Delaunay, and swap
'    diagonal edge of convex quadrilateral if not.
'
    IERR = 0
    X = Vcl(1, I)
    Y = Vcl(2, I)
'
10  'CONTINUE
'
    If (Top <= 0) Then Exit Sub
    T = STACK(Top)
    Top = Top - 1
'
    If (TIL(1, T) = I) Then
        e = 2
        b = TIL(3, T)
    ElseIf (TIL(2, T) = I) Then
        e = 3
        b = TIL(1, T)
    Else
        e = 1
        b = TIL(2, T)
    End If
'
    a = TIL(e, T)
    u = TNBR(e, T)
'
    If (TNBR(1, u) = T) Then
        F = 1
        c = TIL(3, u)
    ElseIf (TNBR(2, u) = T) Then
        F = 2
        c = TIL(1, u)
    Else
        F = 3
        c = TIL(2, u)
    End If
'
    swap = DIAEDG(X, Y, Vcl(1, a), Vcl(2, a), Vcl(1, c), Vcl(2, c), Vcl(1, b), _
        Vcl(2, b))
'
    If (swap = 1) Then
        EM1 = e - 1
        If (EM1 = 0) Then EM1 = 3
        EP1 = e + 1
        If (EP1 = 4) Then EP1 = 1
        fm1 = F - 1
        If (fm1 = 0) Then fm1 = 3
        fp1 = F + 1
        If (fp1 = 4) Then fp1 = 1
        TIL(EP1, T) = c
        TIL(fp1, u) = I
        R = TNBR(EP1, T)
        s = TNBR(fp1, u)
        TNBR(EP1, T) = u
        TNBR(fp1, u) = T
        TNBR(e, T) = s
        TNBR(F, u) = R
'
        If (TNBR(fm1, u) > 0) Then
            Top = Top + 1
            STACK(Top) = u
        End If
'
        If (s > 0) Then
            If (TNBR(1, s) = u) Then
                TNBR(1, s) = T
            ElseIf (TNBR(2, s) = u) Then
                TNBR(2, s) = T
            Else
                TNBR(3, s) = T
            End If
'
            Top = Top + 1
'
            If (Top > MAXST) Then
                IERR = 8
                Exit Sub
            End If
'
            STACK(Top) = T
'
        Else
            If (u = BTRI And fp1 = BEDG) Then
                BTRI = T
                BEDG = e
            End If
'
            L = -(3 * T + e - 1)
            tt = T
            ee = EM1
'
20          'CONTINUE
'
            If (TNBR(ee, tt) > 0) Then
                tt = TNBR(ee, tt)
'
                If (TIL(1, tt) = a) Then
                    ee = 3
                ElseIf (TIL(2, tt) = a) Then
                    ee = 1
                Else
                    ee = 2
                End If
'
                GoTo 20
            End If
'
            TNBR(ee, tt) = L
        End If
'
        If (R > 0) Then
            If (TNBR(1, R) = T) Then
                    TNBR(1, R) = u
            ElseIf (TNBR(2, R) = T) Then
                TNBR(2, R) = u
            Else
                TNBR(3, R) = u
            End If
'
        Else
            If (T = BTRI And EP1 = BEDG) Then
                BTRI = u
                BEDG = F
            End If
'
            L = -(3 * u + F - 1)
            tt = u
            ee = fm1
'
30          'CONTINUE
'
            If (TNBR(ee, tt) > 0) Then
                tt = TNBR(ee, tt)
'
                If (TIL(1, tt) = b) Then
                    ee = 3
                ElseIf (TIL(2, tt) = b) Then
                    ee = 1
                Else
                    ee = 2
                End If
'
                GoTo 30
            End If
'
            TNBR(ee, tt) = L
        End If
'
'        If (msglvl = 4) Then
'            write ( *,600) 2,vcl(1,a),vcl(2,a), &
'                     vcl(1,b),vcl(2,b),x,y,vcl(1,c),vcl(2,c)
'        End If
'600 format (1x,i7,4f15.7/8x,4f15.7)
'
    End If
'
    GoTo 10
'
'
'
'
End Sub
Private Sub VBEDG(ByVal X#, ByVal Y#, Vcl#(), TIL&(), TNBR&(), ByRef LTRI&, _
    ByRef LEDG&, ByRef RTRI&, ByRef REDG&)
'
'******************************************************************************
'
'   VBEDG determines the boundary edges of a 2D triangulation.
'
'
'   Purpose:
'
'    Determine boundary edges of 2D triangulation which are
'    visible from point (X,Y) outside convex hull.
'
'   Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'   Parameters:
'
'    Input, X, Y - 2D point outside convex hull.
'
'    Input, VCL(1:2,1:*) - coordinates of 2D vertices.
'
'    Input, TIL(1:3,1:*) - triangle incidence list.
'
'    Input, TNBR(1:3,1:*) - triangle neighbor list; negative values are
'    used for links of counter clockwise linked list of boundary edges;
'    LINK = -(3*I + J-1) where I, J = triangle, edge index.
'
'    Input, LTRI, LEDG - if LTRI /= 0 then they are assumed to be as defined
'    below and are not changed, else they are updated.  LTRI is
'    the index of boundary triangle to left of leftmost boundary
'    triangle visible from (X,Y).  LEDG is the boundary edge of triangle
'    LTRI to left of leftmost boundary edge visible from (X,Y)
'
'    Input/output, RTRI.  On input, the index of boundary triangle to begin
'    search at.  On output, the index of rightmost boundary triangle
'    visible from (X,Y)
'
'    Input/output, REDG.  On input, the edge of triangle RTRI that is
'    visible from (X,Y).  On output, the edge of triangle RTRI that is
'    visible from (X,Y)
'
    Dim a&, b&, e&, L&, LR&, T&
    Dim LDONE As Boolean
'
'    Find rightmost visible boundary edge using links, then possibly
'    leftmost visible boundary edge using triangle neighbor info.
'
    If (LTRI = 0) Then
        LDONE = False
        LTRI = RTRI
        LEDG = REDG
    Else
        LDONE = True
    End If
'
10  'CONTINUE
'
    L = -TNBR(REDG, RTRI)
    T = L \ 3
    e = (L Mod 3) + 1
    a = TIL(e, T)
'
    If (e <= 2) Then
        b = TIL(e + 1, T)
    Else
        b = TIL(1, T)
    End If
'
    LR = LRLINE(X, Y, Vcl(1, a), Vcl(2, a), Vcl(1, b), Vcl(2, b), 0#)
'
    If (LR > 0) Then
        RTRI = T
        REDG = e
        GoTo 10
    End If
'
    If (LDONE) Then Exit Sub
'
    T = LTRI
    e = LEDG
'
20  'CONTINUE
'
    b = TIL(e, T)
'
    If (e >= 2) Then
        e = e - 1
    Else
        e = 3
    End If
'
    Do While (TNBR(e, T) > 0)
        T = TNBR(e, T)
'
        If (TIL(1, T) = b) Then
            e = 3
        ElseIf (TIL(2, T) = b) Then
            e = 1
        Else
            e = 2
        End If
    Loop
'
    a = TIL(e, T)
    LR = LRLINE(X, Y, Vcl(1, a), Vcl(2, a), Vcl(1, b), Vcl(2, b), 0#)
    If (LR > 0) Then GoTo 20
'
    LTRI = T
    LEDG = e
'
'
'
End Sub
Private Function LRLINE(ByVal XU#, ByVal YU#, _
    ByVal XV1#, ByVal YV1#, ByVal XV2#, ByVal YV2#, ByVal dV#) As Long
'
'******************************************************************************
'
'   LRLINE determines whether a point is left, right, or on a directed line.
'
'
'   Purpose:
'
'    Determine whether a point is to the left of, right of,
'    or on a directed line parallel to a line through given points.
'
'   Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'   Parameters:
'
'    Input, XU, YU, XV1, YV1, XV2, YV2 - vertex coordinates; the directed
'    line is parallel to and at signed distance DV to the
'    left of the directed line from (XV1,YV1) to (XV2,YV2);
'    (XU,YU) is the vertex for which the position
'    relative to the directed line is to be determined.
'
'    Input, DV - signed distance (positive for left).
'
'    Output, LRLINE - +1, 0, or -1 depending on whether (XU,YU) is
'    to the right of, on, or left of the directed line
'    (0 if line degenerates to a point).
'
    Dim Dx#, dxu#, Dy#, dyu#, T#, tol#, tolabs#
'
    tol = 100# * Epsilon '(tol)
    Dx = XV2 - XV1
    Dy = YV2 - YV1
    dxu = XU - XV1
    dyu = YU - YV1
    tolabs = tol * DMAX1(Abs(Dx), Abs(Dy), Abs(dxu), Abs(dyu), Abs(dV))
    T = Dy * dxu - Dx * dyu
'
    If (dV <> 0#) Then
        T = T + dV * Sqr(Dx * Dx + Dy * Dy)
    End If
'
    LRLINE = Int(SIGN(1#, T))
'
    If (Abs(T) <= tolabs) Then
        LRLINE = 0
    End If
'
'
'
End Function
Private Function DLESS(ByVal K&, P#(), ByVal I1&, ByVal J1&, _
    Q#(), ByVal I2&, ByVal J2&) As Boolean
'
'******************************************************************************
'
'   DLESS determines the lexicographically lesser of two double precision values.
'
'
'   Purpose:
'
'    Determine whether P is lexicographically less than Q in
'    floating point arithmetic.
'
'   Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada    T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'   Parameters:
'
'    Input, K - dimension of points.
'
'    Input, P(1:K),Q(1:K) - two K-dimensional double precision points.
'
'    Output, DLESS - .TRUE. if P < Q, .FALSE. otherwise.
'
    Dim I&, cmax#, tol#
'
    tol = 100# * Epsilon '(Tol)
'
    For I = 1 To K
        cmax = DMAX1(Abs(P(I + I1, J1)), Abs(Q(I + I2, J2)))
'
        If (Abs(P(I + I1, J1) - Q(I + I2, J2)) <= tol * cmax Or cmax <= tol) Then
            GoTo 10 'cycle
        End If
'
        If (P(I + I1, J1) < Q(I + I2, J2)) Then
            DLESS = True
        Else
            DLESS = False
        End If
'
        Exit Function
10
    Next I
'
    DLESS = False
'
'
'
End Function
Private Sub DSFTDW(ByVal L&, ByVal u&, ByVal K&, ByVal LDA&, a#(), ByRef MAP&())
'
'******************************************************************************
'
'   DSFTDW does one step of the heap sort algorithm for double precision data.
'
'
'   Purpose:
'
'    Sift A(*,MAP(L)) down a heap of size U.
'
'   Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada    T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'   Parameters:
'
'    Input, L, U - lower and upper index of part of heap.
'
'    Input, K - dimension of points.
'
'    Input, LDA - leading dimension of array A in calling routine.
'
'    Input, A(1:K,1:*), see routine DHPSRT.
'
'    Input/output, MAP(1:*) - see routine DHPSRT.
'
    Dim I&, II&, J&, T&
'
    I = L
    J = 2 * I
    T = MAP(I)
'
    Do
        If (J > u) Then
            Exit Do
        End If
'
        If (J < u) Then
            ' If (DLESS(K, a(1, MAP(J)), a(1, MAP(J + 1)))) Then
            If (DLESS(K, a(), 0, MAP(J), a(), 0, MAP(J + 1))) Then
                J = J + 1
            End If
        End If
'
        ' If (DLESS(K, a(1, MAP(J)), a(1, T))) Then
        If (DLESS(K, a(), 0, MAP(J), a(), 0, T)) Then
            Exit Do
        End If
'
        MAP(I) = MAP(J)
        I = J
        J = 2 * I
    Loop
'
    MAP(I) = T
'
'
'
End Sub
Private Sub DHPSRT(ByVal K&, ByVal N&, ByVal LDA&, a#(), ByRef MAP&())
'
'******************************************************************************
'
'   DHPSRT sorts a list of double precision points in KD.
'
'
'   Purpose:
'
'    Use heapsort to obtain the permutation of N K-dimensional
'    double precision points so that the points are in lexicographic
'    increasing order.
'
'   Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'   Parameters:
'
'    Input, K - dimension of points.
'
'    Input, N - number of points.
'
'    Input, LDA - leading dimension of array A in calling routine; should
'    be >= K.
'
'    Input, A(1:K,1:*) - array of >= N K-D double precision points.
'
'    Input/output, MAP(1:N).  On input, he points of A with indices
'    MAP(1), MAP(2), ..., MAP(N) are to be sorted.  On output, the elements
'    are permuted so that A(*,MAP(1)) <= A(*,MAP(2)) <= ... <= A(*,MAP(N)).
'
    Dim I&, T&
'
    For I = N \ 2 To 1 Step -1
        Call DSFTDW(I, N, K, LDA, a, MAP)
    Next I
'
    For I = N To 2 Step -1
        T = MAP(1)
        MAP(1) = MAP(I)
        MAP(I) = T
        Call DSFTDW(1, I - 1, K, LDA, a, MAP)
    Next I
'
'
'
End Sub
Public Sub DTRIS2(ByVal NPT&, Vcl#(), Ind&(), _
    ByRef NTRI&, ByRef TIL&(), ByRef TNBR&(), ByRef IERR&)
'
'******************************************************************************
'
'   DTRIS2 constructs the Delaunay triangulation of vertices in 2D.
'
'
'   Purpose:
'
'    Construct Delaunay triangulation of 2D vertices using
'    incremental approach and diagonal edge swaps.    Vertices are
'    first sorted in lexicographically increasing (x,y) order, and
'    then are inserted one at a time from outside the convex hull.
'
'   Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada    T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'   Parameters:
'
'    Input, NPT - number of 2D points (vertices).
'
'    Input, MAXST - maximum size available for STACK array; should be about
'    NPT to be safe, but MAX(10,2*LOG2(NPT)) usually enough.
'
'    Input, VCL(1:2,1:*) - coordinates of 2D vertices.
'
'    Input/output, IND(1:NPT) - indices in VCL of vertices to be triangulated.
'    On output, permuted due to sorting.
'
'    Output, NTRI - number of triangles in triangulation; equal to
'    2*NPT - NB - 2 where NB = number of boundary vertices.
'
'    Output, TIL(1:3,1:NTRI) - triangle incidence list; elements are indices.
'    of VCL; vertices of triangles are in counter clockwise order.
'
'    Output, TNBR(1:3,1:NTRI) - triangle neighbor list; positive elements
'    are indices of TIL; negative elements are used for links
'    of counter clockwise linked list of boundary edges;
'    LINK = -(3*I + J-1) where I, J = triangle, edge index;
'    TNBR(J,I) refers to the neighbor along edge from vertex
'    J to J+1 (mod 3).
'
'    Workspace, STACK(1:MAXST) - used for stack of triangles for which
'    circumcircle test must be made.
'
'    Output, integer IERR, error flag, which is zero unless an error occurred.
'
    Dim cmax#, tol#
    Dim I&, J&, e&, L&, LEDG&, LR&, LTRI&, M&, M1&, M2&, N&, MAXST&, REDG&
    Dim RTRI&, T&, Top&
'
    MAXST = NPT
    ReDim STACK&(1 To MAXST)
'
    Const msglvl& = 0
'
    IERR = 0
    tol = 100# * Epsilon '(tol)
'
'    Sort vertices by increasing (x,y).
'
    Call DHPSRT(2, NPT, 2, Vcl, Ind)
'
'    Check that the data is not degenerate.
'
    M1 = Ind(1)
'
    For I = 2 To NPT
        M = M1
        M1 = Ind(I)
'
        For J = 1 To 2
            cmax = DMAX1(Abs(Vcl(J, M)), Abs(Vcl(J, M1)))
            If (Abs(Vcl(J, M) - Vcl(J, M1)) > tol * cmax And cmax > tol) Then
                GoTo 20
            End If
        Next J
'
        IERR = 224
        Exit Sub
'
20      'CONTINUE
    Next I
'
'    Staring with points M1 and M2, find the first point M that is
'    "reasonably" non-colinear.
'
    M1 = Ind(1)
    M2 = Ind(2)
    J = 3
'
    Do
        If (J > NPT) Then
            IERR = 225
            Exit Sub
        End If
'
        M = Ind(J)
        LR = LRLINE(Vcl(1, M), Vcl(2, M), Vcl(1, M1), Vcl(2, M1), _
                    Vcl(1, M2), Vcl(2, M2), 0#)
'
        If (LR <> 0) Then
            Exit Do
        End If
'
        J = J + 1
    Loop
'
'    Set up the initial triangle information for M1, M2 and M (and any
'    in-between points we may have skipped over while searching for M.
'
    NTRI = J - 2
'
    If (LR = -1) Then
        TIL(1, 1) = M1
        TIL(2, 1) = M2
        TIL(3, 1) = M
        TNBR(3, 1) = -3
'
        For I = 2 To NTRI
            M1 = M2
            M2 = Ind(I + 1)
            TIL(1, I) = M1
            TIL(2, I) = M2
            TIL(3, I) = M
            TNBR(1, I - 1) = -3 * I
            TNBR(2, I - 1) = I
            TNBR(3, I) = I - 1
        Next I
'
        TNBR(1, NTRI) = -3 * NTRI - 1
        TNBR(2, NTRI) = -5
        LEDG = 2
        LTRI = NTRI
'
    Else
        TIL(1, 1) = M2
        TIL(2, 1) = M1
        TIL(3, 1) = M
        TNBR(1, 1) = -4
'
        For I = 2 To NTRI
            M1 = M2
            M2 = Ind(I + 1)
            TIL(1, I) = M2
            TIL(2, I) = M1
            TIL(3, I) = M
            TNBR(3, I - 1) = I
            TNBR(1, I) = -3 * I - 3
            TNBR(2, I) = I - 1
        Next I
'
        TNBR(3, NTRI) = -3 * NTRI
        TNBR(2, 1) = -3 * NTRI - 2
        LEDG = 2
        LTRI = 1
    End If
'
'    If (msglvl = 4) Then
'       M2 = IND(1)
'       write ( *,'(i7,4f15.7)') 1,vcl(1,m2),vcl(2,m2),vcl(1,m),vcl(2,m)
'       do i = 2,j-1
'           M1 = M2
'           M2 = IND(I)
'           write ( *,'(i7,4f15.7)') 1,vcl(1,m1),vcl(2,m1),vcl(1,m2),vcl(2,m2)
'           write ( *,'(i7,4f15.7)') 1,vcl(1,m2),vcl(2,m2),vcl(1,m),vcl(2,m)
'       end do
'    End If
'
'    Insert vertices one at a time from outside convex hull, determine
'    visible boundary edges, and apply diagonal edge swaps until
'    Delaunay triangulation of vertices (so far) is obtained.
'
    Top = 0
'
    For I = J + 1 To NPT
'    If (msglvl = 4) Then
'        write ( *,'(i7,4f15.7)') i
'    End If
'
        M = Ind(I)
        M1 = TIL(LEDG, LTRI)
'
        If (LEDG <= 2) Then
            M2 = TIL(LEDG + 1, LTRI)
        Else
            M2 = TIL(1, LTRI)
        End If
'
        LR = LRLINE(Vcl(1, M), Vcl(2, M), Vcl(1, M1), Vcl(2, M1), _
                    Vcl(1, M2), Vcl(2, M2), 0#)
'
        If (LR > 0) Then
            RTRI = LTRI
            REDG = LEDG
            LTRI = 0
        Else
            L = -TNBR(LEDG, LTRI)
            RTRI = L \ 3
            REDG = (L Mod 3) + 1
        End If
'
        Call VBEDG(Vcl(1, M), Vcl(2, M), Vcl(), TIL(), TNBR(), LTRI, LEDG, RTRI, REDG)
        N = NTRI + 1
        L = -TNBR(LEDG, LTRI)
'
        Do
            T = L \ 3
            e = (L Mod 3) + 1
            L = -TNBR(e, T)
            M2 = TIL(e, T)
'
            If (e <= 2) Then
                M1 = TIL(e + 1, T)
            Else
                M1 = TIL(1, T)
            End If
'
            NTRI = NTRI + 1
            TNBR(e, T) = NTRI
            TIL(1, NTRI) = M1
            TIL(2, NTRI) = M2
            TIL(3, NTRI) = M
            TNBR(1, NTRI) = T
            TNBR(2, NTRI) = NTRI - 1
            TNBR(3, NTRI) = NTRI + 1
            Top = Top + 1
'
            If (Top > MAXST) Then
                IERR = 8
                Exit Sub
            End If
'
            STACK(Top) = NTRI
    
'           If (msglvl = 4) Then
'               write ( *,'(i7,4f15.7)') 1,vcl(1,m),vcl(2,m), vcl(1,m2),vcl(2,m2)
'           End If
'
            If (T = RTRI And e = REDG) Then
                Exit Do
            End If
        Loop
'
'       if (msglvl = 4) then
'           write ( *,'(i7,4f15.7)') 1,vcl(1,m),vcl(2,m), vcl(1,m1),vcl(2,m1)
'       End If
'
        TNBR(LEDG, LTRI) = -3 * N - 1
        TNBR(2, N) = -3 * NTRI - 2
        TNBR(3, NTRI) = -L
        LTRI = N
        LEDG = 2
        Call SWAPEC(M, Top, MAXST, LTRI, LEDG, Vcl, TIL, TNBR, STACK, IERR)
'
        If (IERR <> 0) Then
            Exit Sub
        End If
    Next I
'
'    If (msglvl = 4) Then
'       write ( *, '(i7,4f15.7)' ) npt+1
'    End If
'
'
'
End Sub
